/* strong_hits.c -- find matches in an alignment that meet specified 
 *                  constraints for length and strength
 *
 * Please see the PipTools Program Reference (.../docs/piptools_ref.html)
 * for a description of what this program does and how to use it.
 *
 * This program was suggested by John Pedan.
 *
 * $Revision: 1.9 $$Date: 2001/10/31 20:12:45 $
 */

#include "lav.h"
#include "exons.h"
#include "feat.h"
#include "util.h"

#define SYNTAX  "align_file -len min_length -pct min_ident " \
                  "[-gap max_gap_size -step max_ident_step] " \
                  "[-overlaps interval_file [-inseq2] [-exons-only]] " \
                  "> output_file"

enum { BUFSIZE = 1000 };

static void get_args (int argc, char **argv, char **alignfile,
                      int *min_len, int *min_pct, int *max_gap, int *max_step,
                      char **intfile, bool *inseq2, bool *exonly);
static int get_intervals (char *intfile, int **start, int **end);
static int prec_gap (Segment *seg, int i);
static void print_group (Segment *seg, int first, int i,
		               int totscore, int totlen, int max_gap, char **printed,
		               int *start, int *end, char **gname, int nint,
		               bool exonly, bool inseq2);
static void print_overlaps (int a0, int a1, int *start, int *end,
                            char **gname, int nint, bool exonly);

int main (int argc, char **argv) {
	char *alignfile = NULL;
	int min_len = 0;
	int min_pct = 0;
	int max_gap = 0;
	int max_step = 0;
	char *intfile = NULL;
	bool inseq2 = 0;
	bool exonly = 0;
	Segment *seg;                      /* array of alignment info */
	int size;                          /* allocated size of seg array */
	int *start = NULL, *end = NULL;    /* intervals (parallel arrays) */
	char **gname = NULL;               /* more gene/exon info */
	char **ename = NULL;
	int *dir = NULL;
	int nint;                          /* number of intervals */
	int first = -1;                    /* index of group's first segment */
	int totscore = 0;                  /* sum of (score * len) for group */
	int totlen = 0;                    /* total length of group, excl. gaps */
	char type;
	char *printed = NULL;
	int i, score, len, nhits = 0;

	/* get arguments */
	set_argv0 (argv[0]);
	get_args (argc, argv, &alignfile, &min_len, &min_pct, &max_gap, &max_step,
	                      &intfile, &inseq2, &exonly);
	if (!alignfile || min_len < 1 || min_pct < 1 || min_pct > 100 ||
			max_gap < 0 || max_step < 0 || XOR (max_gap, max_step) ||
			((inseq2 || exonly) && !intfile)) {
		fatal (usage (SYNTAX));
	}

	/* read alignment file into Segment array */
	size = get_align (alignfile, &seg);
	if (has_multi_contig (seg) && inseq2)
		fatalf ("2nd-sequence overlap intervals do not work with " \
		        "multiple contigs.");

	/* read interval file into parallel arrays */
	if (exonly)
		nint = get_exons (intfile, &start, &end, &dir, &gname, &ename, -1);
	else
		nint = get_intervals (intfile, &start, &end);

	/* print title */
	if (max_gap) {
		printf ("Groups of segments in %s\n", alignfile);
		printf ("with overall length >= %d bp", min_len);
		printf (" and individual identity >= %d%%,\n", min_pct);
		printf ("where gaps are <= %d bp, and identity steps are <= %d%%",
		        max_gap, max_step);
	}
	else {
		printf ("Gap-free segments in %s\n", alignfile);
		printf ("with length >= %d bp", min_len);
		printf (" and identity >= %d%%", min_pct);
	}
	if (intfile) {
		printf (",\nshowing %s from %s that overlap them in sequence %d",
		        (exonly ? "exons" : "intervals"), intfile, (inseq2 ? 2 : 1));
	}
	printf (":\n\n");

	/* find strong matches */
	for (i = 0; (type = seg[i].type) != 'z'; i++) {
		if (type == 'l') {
			score = seg[i].score;
			len = abs (seg[i].x1 - seg[i].x0) + 1;

			if (first < 0) {
				if (score >= min_pct) {
					/* start new group */
					first = i;
					totscore = score * len;
					totlen = len;
				}
			}
			else if (same_local (seg, first, i) &&      /* i > first >= 0 */
					prec_gap (seg, i) <= max_gap &&
					score >= min_pct &&
					abs (score - seg[i-1].score) <= max_step) {
				/* continue current group */
				totscore += score * len;
				totlen += len;    /* don't include gap */
			}
			else {
				/* finish current group */
				if (totlen >= min_len) {
					print_group (seg, first, i-1, totscore, totlen,
					             max_gap, &printed, start, end, gname, nint,
					             exonly, inseq2);
					nhits++;
				}
				first = -1;
				i--;    /* re-examine this segment as possible new first */
			}
		}
	}
	/* finish last group */
	if (first >= 0 && totlen >= min_len) {
		print_group (seg, first, i-1, totscore, totlen, max_gap, &printed,
		             start, end, gname, nint, exonly, inseq2);
		nhits++;
	}
	if (!nhits)
		printf ("No aligned segments matched these criteria.\n");
	else
		printf ("\n%d hit%s.\n", nhits, (nhits == 1 ? "" : "s"));

	/* clean up */
	free_align (seg, size);
	if (exonly)
		free_exons (start, end, dir, gname, ename, nint);
	else {
		free (start);
		free (end);
	}
	return 0;
}

/* Retrieve command-line arguments */
static void get_args (int argc, char **argv, char **alignfile,
                      int *min_len, int *min_pct, int *max_gap, int *max_step,
                      char **intfile, bool *inseq2, bool *exonly) {
	int i;

	for (i = 1; i < argc; i++) {
		if (strsame (argv[i], "-len")) {
			if (!(*min_len = atoi (optval (argc, argv, &i, SYNTAX))))
				fatal ("Invalid min_length value.");   /* catch non-ints */
		}
		else if (strsame (argv[i], "-pct")) {
			if (!(*min_pct = atoi (optval (argc, argv, &i, SYNTAX))))
				fatal ("Invalid min_ident value.");
		}
		else if (strsame (argv[i], "-gap")) {
			if (!(*max_gap = atoi (optval (argc, argv, &i, SYNTAX))))
				fatal ("Invalid max_gap_size value.");
		}
		else if (strsame (argv[i], "-step")) {
			if (!(*max_step = atoi (optval (argc, argv, &i, SYNTAX))))
				fatal ("Invalid max_ident_step value.");
		}
		else if (strsame (argv[i], "-overlaps"))
			*intfile = optval (argc, argv, &i, SYNTAX);
		else if (strsame (argv[i], "-inseq2"))
			*inseq2 = 1;
		else if (strsame (argv[i], "-exons-only"))
			*exonly = 1;
		else if (!*alignfile)
			*alignfile = argv[i];
		else
			fatal (usage (SYNTAX));
	}
}

/* Read interval file into parallel arrays */
static int get_intervals (char *intfile, int **start, int **end) {
	FILE *fp;
	char buf [BUFSIZE];
	char head [BUFSIZE], mid [BUFSIZE], *tail;   /* non-addr parts of buf */
	int a1, a2;                                  /* interval endpoints */
	char *p;
	int i;

	if (!intfile)
		return 0;

	/* allocate parallel arrays */
	fp = ckopen (intfile, "r");
	i = 0;
	while (fgets (buf, BUFSIZE, fp)) {
		p = skip_ws (buf);
		if (*p && !starts (p, "#"))
			i++;
	}
	*start = (int *) ckalloc (i * sizeof (int));
	*end   = (int *) ckalloc (i * sizeof (int));

	/* read intervals */
	rewind (fp);
	i = 0;
	while (fgets (buf, BUFSIZE, fp)) {
		if (feat_parse_line (buf, head, &a1, mid, &a2, &tail)) {
			(*start)[i] = a1;
			(*end)[i]   = a2;
			i++;
		}
	}

	fclose (fp);
	return i;
}

/* Compute size of the gap preceding a particular segment */
static int prec_gap (Segment *seg, int i) {
	int xgap, ygap;

	assert (i > 0 && seg[i-1].type == 'l' && seg[i].type == 'l');
	xgap = abs (seg[i].x0 - seg[i-1].x1) - 1;
	ygap = abs (seg[i].y0 - seg[i-1].y1) - 1;
	assert (xgap <= abs (seg[i].x1 - seg[i-1].x0) - 1);
	assert (ygap <= abs (seg[i].y1 - seg[i-1].y0) - 1);
	assert (!xgap || !ygap);
	return xgap + ygap;
}

/* Print output line for group */
static void print_group (Segment *seg, int first, int i,
		               int totscore, int totlen, int max_gap, char **printed,
		               int *start, int *end, char **gname, int nint,
		               bool exonly, bool inseq2) {
	char *hdr2;
	int x0, x1, y0, y1;
	int j, last;

	/* print seq2 header */
	if (!(hdr2 = get_hdr2 (seg, first)))
		fatalf ("Missing sequence header for segment %d.", first);
	if (!strsame (hdr2, *printed)) {
		printf ("%s\n", hdr2);
		*printed = hdr2;
	}

	/* get group endpoints */
	for (j = first; j <= i && seg[j].type == 'l'; j++)  /* drop trailing s,a */
		;
	last = j-1;
	x0 = seg[first].x0;
	x1 = seg[last].x1;
	y0 = seg[first].y0;
	y1 = seg[last].y1;
	assert (abs (x0 - x1) >= abs (seg[last].x0 - seg[first].x1));
	assert (abs (y0 - y1) >= abs (seg[last].y0 - seg[first].y1));

	/* print output line */
	printf ("    %d-%d <--> %d-%d %.0f%% (%d nt", x0, x1, y0, y1,
	        totscore * 1.0 / totlen, totlen);
	if (max_gap)
		printf ("; %d seg", last - first + 1);
	printf (")");
	if (seg[first].reverse)
		printf (" (reverse complement)");
	if (!inseq2)
		print_overlaps (x0, x1, start, end, gname, nint, exonly);
	else
		print_overlaps (y0, y1, start, end, gname, nint, exonly);
	putchar ('\n');
}

/* Print overlaps from interval file */
static void print_overlaps (int a0, int a1, int *start, int *end,
                            char **gname, int nint, bool exonly) {
	int j, count = 0;

	for (j = 0; j < nint; j++) {
		if (exonly && gname[j]) {    /* skip gene and cds */
			j++;
			continue;
		}
		if (overlaps (a0, a1, start[j], end[j])) {
			printf ("%s %d-%d", (!count ? "; overlaps" : ","),
			        start[j], end[j]);
			count++;
		}
	}
}

